/* Angophora Electronics temperature and humidity sensor 2009-12-10
   http://sites.google.com/site/angophora
   Microcontroller           ATMEGA8
   Humidity sensor           HONEYWELL HCH-1000-002
   Temperature sensor        LM335AZ
   Crystal                   3.6864-18  MHz
*/
#include <avr/io.h> 
#include <string.h>
#define F_CPU 3686400  // 3.6864 MHz
#include <util/delay.h>
#include <stdlib.h>
#include <avr/interrupt.h>
#define BAUD 115200
#define MYUBRR F_CPU/16/BAUD-1


void initalize_USART( unsigned int ubrr){
   UBRRH = (unsigned char)(ubrr>>8);       // Set baud rate high byte
   UBRRL = (unsigned char)ubrr;            // Set baud rate low byte
   UCSRB = (1<<RXEN)|(1<<TXEN);               // Enable receiver and transmitter
   UCSRC = (1<<URSEL)|(0<<USBS)|(3<<UCSZ0);       // Set frame format: 8data, 1stop bit
}


void send_byte(unsigned char byt){         // this function sends a single ascii character to the serial interface
    while (!(UCSRA & (1<<UDRE)));           // Wait while NOT ready to transmit
	UDR = byt;                           // send byte
}


void send_string(char * str) {         //  this function sends a string to the serial interface using a pointer
    while(*str) {                     // while not end of string
        send_byte(*str++);         // iterate through the string sending byte by byte
    }
} 


unsigned char read_byte(void) {          // this function will be used for polled reading
	while(!(UCSRA & (1<<RXC))); 		// Wait while data is NOT available to read
	return UDR;
}



void initalize_ADC( void ){          // this function sets up the analogue to digital converter
    ADMUX=(1<<REFS0);              // set Aref = AVcc;
    ADCSRA=(1<<ADEN)|(7<<ADPS0);       //  Enable ADC and set prescaling
}


int read_ADC(char ch){               // this function reads the analogue to digital converter
    int n, av=64;                       // number of times to average the adc
    unsigned int average_adc = 0;            // variable to store the adc sums       
    ADMUX |= (ch & 0x07);            //force to channel 0-7 and set ADC channel
	for(n=0; n<av; n++){            // this function reads the adc multiple times to average electrical noise
       ADCSRA |= (1<<ADSC);                  //Start Single conversion
       while( !(ADCSRA & (1<<ADIF)) );     //Wait for conversion to complete
       ADCSRA |= (1<<ADIF);                       //Clear ADIF by writing one to it
	   average_adc += ADC;              // sum all the adc readings
	}
	average_adc /= av;             // divide the sum of adc readings by the number of adc readings to get the average
    return (int) average_adc;      //  the (int) here converts the variable to a signed integer 
}


int read_temperature( void ){          // this function reads the LM335AZ sensor
    int  adc_value,temperature_value;                     
    PORTC |= (1<<PC1);                           //  switch on temperature sensor
	_delay_ms(80);                                   // wait for sensor to stabilize
    adc_value = read_ADC(0);                       // read adc channel 0
	PORTC &= ~(1<<PC1);                                    // switch off temperature sensor
    temperature_value = 0.48828125 * adc_value - 273;         // adc_value * VCC in mV / 10 bits / 10 mV  - 273 degrees Kelvin
    return temperature_value;            
}



void zero_counter( void ){         // this function sets the 16 bit counter to zero
    unsigned char sreg;
    sreg = SREG;            // save global interrupt flag
	cli();             // disable interrupts
    TCNT1 = 0;            // set counter register to zero 
    SREG = sreg;           // restore global interrupt flag
	sei();              // enable interupts
}


unsigned int read_counter( void ){        // this function reads the 16 bit counter
    unsigned char sreg;
    unsigned int count;
    sreg = SREG;          // save global interrupt flag
    cli();                 // disable interrupts
    count = TCNT1;         // read counter register
    SREG = sreg;        // restore global interrupt flag
	sei();               // enable interupts
    return count;
}



int read_humidity( void ){
    unsigned int counter_value;
	int humidity_value;
	float m = 0.147929;               // slope
	float a = 427.515;                //  offset  

    zero_counter();                            //  set 16 bit counter to zero
	PORTB &= ~(1<<PB1);                            //  enable humidity oscillator
	TCCR1B |= ((0<<CS12)|(1<<CS11)|(0<<CS10));      // set timer 1 to 0 1 0 clkI/O/8 (From prescaler) internal clock
    while((!(PINB & 1)) & (TCNT1<65000));        // wait for sensor output to go high, & TNCT<65000 in case output never goes high       
    TCCR1B &= ~((1<<CS12)|(1<<CS11)|(1<<CS10));   // switch off timer 1 , 0 0 0 No clock source (Timer/Counter stopped).
    PORTB |= (1<<PB1);                              // reset humidity oscillator
    counter_value = read_counter();                            //  read the counter
    humidity_value = (m * counter_value) - a;      // convert counter_value into relative humidity
//	return counter_value;                                // use only for calibrating the sensor
    return humidity_value;                       // normal return value
}


void delay_minutes( int delay ){
    int t;
    for (t = 0; t < (delay * 60); t++){
    _delay_ms(1000);
	}
}



int main( void ){
	char temperature_str[8];
	char humidity_str[8];
	int temperature;
	int humidity;

    initalize_USART( MYUBRR );     // initalize the USART for serial communication
    initalize_ADC();           //initialize the ADC for the temperature sensor

    DDRC |= (1<<DDC1);        //  set PC1 as output to power the temperature sensor
	DDRB |= (1<<DDB1);        //  set PB1 as output to power the humidity sensor
	PORTB |= (1<<PB1);        //  reset humidity oscillator

    while( 1 ){                               // loop forever
	    temperature = read_temperature();         // read temperature	
		humidity = read_humidity();              //read humidity

    	itoa(temperature,temperature_str,10);        // convert temperature into ascii string
	    itoa(humidity,humidity_str,10);           // convert humidity into ascii string

		send_string(temperature_str);          // send data to serial port
		send_string(" degC");             // send data to serial port
        send_string("  ");            // send data to serial port

		send_string(humidity_str);          // send data to serial port
		send_string(" %RH");          // send data to serial port
        send_string("\r\n");            // send data to serial port

		_delay_ms(1000); 
    }
    return 0;
}













 
